@@ -1,6 +1,6 @@
 /* evalstring.c - evaluate a string as one or more shell commands. */
 
-/* Copyright (C) 1996-2021 Free Software Foundation, Inc.
+/* Copyright (C) 1996-2023 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -63,30 +63,65 @@ extern int errno;
 
 int parse_and_execute_level = 0;
 
-static int cat_file PARAMS((REDIRECT *));
+static int cat_file (REDIRECT *);
 
 #define PE_TAG "parse_and_execute top"
 #define PS_TAG "parse_string top"
 
 #if defined (HISTORY)
 static void
-set_history_remembering ()
+uw_set_history_remembering (void *ignore)
 {
   remember_on_history = enable_history_list;
 }
+
 #endif
 
 static void
-restore_lastcom (x)
-     char *x;
+uw_restore_lastcom (void *x)
 {
   FREE (the_printed_command_except_trap);
   the_printed_command_except_trap = x;
 }
 
+static void
+uw_set_current_prompt_level (void *x)
+{
+  set_current_prompt_level ((intptr_t) x);
+}
+
+static void
+uw_pop_stream (void *x)
+{
+  pop_stream ();
+}
+
+static void
+uw_parser_restore_alias (void *x)
+{
+  parser_restore_alias ();
+}
+
 int
-should_suppress_fork (command)
-     COMMAND *command;
+should_optimize_fork (COMMAND *command, int subshell)
+{
+  return (running_trap == 0 &&
+      command->type == cm_simple &&
+      signal_is_trapped (EXIT_TRAP) == 0 &&
+      signal_is_trapped (ERROR_TRAP) == 0 &&
+      any_signals_trapped () < 0 &&
+      (subshell || (command->redirects == 0 && command->value.Simple->redirects == 0)) &&
+      ((command->flags & CMD_TIME_PIPELINE) == 0) &&
+      ((command->flags & CMD_INVERT_RETURN) == 0));
+}
+
+/* This has extra tests to account for STARTUP_STATE == 2, which is for
+   -c command but has been extended to command and process substitution
+   (basically any time you call parse_and_execute in a subshell). */
+/* ssh_reading_startup_files will never be non-zero unless someone goes
+   and uncomments SSH_SOURCE_BASHRC in config-top.h */
+int
+should_suppress_fork (COMMAND *command)
 {
   int subshell;
 
@@ -94,19 +129,13 @@ should_suppress_fork (command)
   return (startup_state == 2 && parse_and_execute_level == 1 &&
 	  *bash_input.location.string == '\0' &&
 	  parser_expanding_alias () == 0 &&
-	  running_trap == 0 &&
-	  command->type == cm_simple &&
-	  signal_is_trapped (EXIT_TRAP) == 0 &&
-	  signal_is_trapped (ERROR_TRAP) == 0 &&
-	  any_signals_trapped () < 0 &&
-	  (subshell || (command->redirects == 0 && command->value.Simple->redirects == 0)) &&
-	  ((command->flags & CMD_TIME_PIPELINE) == 0) &&
-	  ((command->flags & CMD_INVERT_RETURN) == 0));
+	  job_control_active_p () == 0 &&
+	  ssh_reading_startup_files == 0 &&
+	  should_optimize_fork (command, subshell));
 }
 
 int
-can_optimize_connection (command)
-     COMMAND *command;
+can_optimize_connection (COMMAND *command)
 {
   return (*bash_input.location.string == '\0' &&
 	  parser_expanding_alias () == 0 &&
@@ -115,13 +144,13 @@ can_optimize_connection (command)
 }
 
 void
-optimize_fork (command)
-     COMMAND *command;
+optimize_connection_fork (COMMAND *command)
 {
   if (command->type == cm_connection &&
       (command->value.Connection->connector == AND_AND || command->value.Connection->connector == OR_OR || command->value.Connection->connector == ';') &&
       (command->value.Connection->second->flags & CMD_TRY_OPTIMIZING) &&
-      should_suppress_fork (command->value.Connection->second))
+      (should_suppress_fork (command->value.Connection->second) ||
+      ((subshell_environment & SUBSHELL_PAREN) && should_optimize_fork (command->value.Connection->second, 0))))
     {
       command->value.Connection->second->flags |= CMD_NO_FORK;
       command->value.Connection->second->value.Simple->flags |= CMD_NO_FORK;
@@ -129,29 +158,25 @@ optimize_fork (command)
 }
 
 void
-optimize_subshell_command (command)
-     COMMAND *command;
+optimize_subshell_command (COMMAND *command)
 {
-  if (running_trap == 0 &&
-      command->type == cm_simple &&
-      signal_is_trapped (EXIT_TRAP) == 0 &&
-      signal_is_trapped (ERROR_TRAP) == 0 &&
-      any_signals_trapped () < 0 &&
-      command->redirects == 0 && command->value.Simple->redirects == 0 &&
-      ((command->flags & CMD_TIME_PIPELINE) == 0) &&
-      ((command->flags & CMD_INVERT_RETURN) == 0))
+  if (should_optimize_fork (command, 0))
     {
       command->flags |= CMD_NO_FORK;
       command->value.Simple->flags |= CMD_NO_FORK;
     }
   else if (command->type == cm_connection &&
-	   (command->value.Connection->connector == AND_AND || command->value.Connection->connector == OR_OR))
-    optimize_subshell_command (command->value.Connection->second);
+	   (command->value.Connection->connector == AND_AND || command->value.Connection->connector == OR_OR || command->value.Connection->connector == ';') &&
+	   command->value.Connection->second->type == cm_simple &&
+	   parser_expanding_alias () == 0)
+    {	   
+      command->value.Connection->second->flags |= CMD_TRY_OPTIMIZING;
+      command->value.Connection->second->value.Simple->flags |= CMD_TRY_OPTIMIZING;
+    }
 }
 
 void
-optimize_shell_function (command)
-     COMMAND *command;
+optimize_shell_function (COMMAND *command)
 {
   COMMAND *fc;
 
@@ -170,8 +195,7 @@ optimize_shell_function (command)
 }
 
 int
-can_optimize_cat_file (command)
-     COMMAND *command;
+can_optimize_cat_file (COMMAND *command)
 {
   return (command->type == cm_simple && !command->redirects &&
 	    (command->flags & CMD_TIME_PIPELINE) == 0 &&
@@ -184,8 +208,7 @@ can_optimize_cat_file (command)
 
 /* How to force parse_and_execute () to clean up after itself. */
 void
-parse_and_execute_cleanup (old_running_trap)
-     int old_running_trap;
+parse_and_execute_cleanup (int old_running_trap)
 {
   if (running_trap > 0)
     {
@@ -205,10 +228,7 @@ parse_and_execute_cleanup (old_running_trap)
 }
 
 static void
-parse_prologue (string, flags, tag)
-     char *string;
-     int flags;
-     char *tag;
+parse_prologue (char *string, int flags, char *tag)
 {
   char *orig_string, *lastcom;
   int x;
@@ -222,14 +242,15 @@ parse_prologue (string, flags, tag)
   unwind_protect_int (line_number);
   unwind_protect_int (line_number_for_err_trap);
   unwind_protect_int (loop_level);
-  unwind_protect_int (executing_list);
+  unwind_protect_int (interrupt_execution);
   unwind_protect_int (comsub_ignore_return);
+  unwind_protect_int (builtin_ignoring_errexit);
   if (flags & (SEVAL_NONINT|SEVAL_INTERACT))
     unwind_protect_int (interactive);
 
 #if defined (HISTORY)
   if (parse_and_execute_level == 0)
-    add_unwind_protect (set_history_remembering, (char *)NULL);
+    add_unwind_protect (uw_set_history_remembering, (char *)NULL);
   else
     unwind_protect_int (remember_on_history);	/* can be used in scripts */
 #  if defined (BANG_HISTORY)
@@ -240,18 +261,18 @@ parse_prologue (string, flags, tag)
   if (interactive_shell)
     {
       x = get_current_prompt_level ();
-      add_unwind_protect (set_current_prompt_level, x);
+      add_unwind_protect (uw_set_current_prompt_level, (void *) (intptr_t) x);
     }
 
   if (the_printed_command_except_trap)
     {
       lastcom = savestring (the_printed_command_except_trap);
-      add_unwind_protect (restore_lastcom, lastcom);
+      add_unwind_protect (uw_restore_lastcom, lastcom);
     }
 
-  add_unwind_protect (pop_stream, (char *)NULL);
+  add_unwind_protect (uw_pop_stream, (char *)NULL);
   if (parser_expanding_alias ())
-    add_unwind_protect (parser_restore_alias, (char *)NULL);
+    add_unwind_protect (uw_parser_restore_alias, (char *)NULL);
 
   if (orig_string && ((flags & SEVAL_NOFREE) == 0))
     add_unwind_protect (xfree, orig_string);
@@ -280,15 +301,13 @@ parse_prologue (string, flags, tag)
    	(flags & SEVAL_NOFREE) -> don't free STRING when finished
    	(flags & SEVAL_RESETLINE) -> reset line_number to 1
    	(flags & SEVAL_NOHISTEXP) -> history_expansion_inhibited -> 1
+   	(flags & SEVAL_NOOPTIMIZE) -> don't try to turn on optimizing flags
 */
 
 int
-parse_and_execute (string, from_file, flags)
-     char *string;
-     const char *from_file;
-     int flags;
+parse_and_execute (char *string, const char *from_file, int flags)
 {
-  int code, lreset;
+  int code, lreset, ignore_return;
   volatile int should_jump_to_top_level, last_result;
   COMMAND *volatile command;
   volatile sigset_t pe_sigmask;
@@ -298,6 +317,8 @@ parse_and_execute (string, from_file, flags)
   parse_and_execute_level++;
 
   lreset = flags & SEVAL_RESETLINE;
+  ignore_return = (this_shell_builtin == eval_builtin || this_shell_builtin == source_builtin) &&
+		  builtin_ignoring_errexit;
 
 #if defined (HAVE_POSIX_SIGNALS)
   /* If we longjmp and are going to go on, use this to restore signal mask */
@@ -353,12 +374,14 @@ parse_and_execute (string, from_file, flags)
 		 these circumstances.  Don't bother with cleanup here because
 		 we don't want to run the function execution cleanup stuff
 		 that will cause pop_context and other functions to run.
+		 We call reset_local_contexts() instead, which just frees
+		 context memory.
 		 XXX - change that if we want the function context to be
 		 unwound. */
 	      if (exit_immediately_on_error && variable_context)
 	        {
 	          discard_unwind_frame ("pe_dispose");
-		  variable_context = 0;	/* not in a function */
+	          reset_local_contexts (); /* not in a function */
 	        }
 	      should_jump_to_top_level = 1;
 	      goto out;
@@ -395,6 +418,7 @@ parse_and_execute (string, from_file, flags)
 		run_unwind_frame ("pe_dispose");
 	      last_result = last_command_exit_value = EXECUTION_FAILURE; /* XXX */
 	      set_pipestatus_from_exit (last_command_exit_value);
+	      
 	      if (subshell_environment)
 		{
 		  should_jump_to_top_level = 1;
@@ -416,9 +440,11 @@ parse_and_execute (string, from_file, flags)
 	      break;
 	    }
 	}
-	  
+
       if (parse_command () == 0)
 	{
+	  int local_expalias, local_alflag;
+
 	  if ((flags & SEVAL_PARSEONLY) || (interactive_shell == 0 && read_but_dont_execute))
 	    {
 	      last_result = EXECUTION_SUCCESS;
@@ -446,6 +472,8 @@ parse_and_execute (string, from_file, flags)
 		      should_jump_to_top_level = 0;
 		      last_result = last_command_exit_value = EX_BADUSAGE;
 		      set_pipestatus_from_exit (last_command_exit_value);
+		      dispose_command (command);
+		      global_command = (COMMAND *)NULL;
 		      reset_parser ();
 		      break;
 		    }
@@ -453,14 +481,17 @@ parse_and_execute (string, from_file, flags)
 
 	      bitmap = new_fd_bitmap (FD_BITMAP_SIZE);
 	      begin_unwind_frame ("pe_dispose");
-	      add_unwind_protect (dispose_fd_bitmap, bitmap);
-	      add_unwind_protect (dispose_command, command);	/* XXX */
+	      add_unwind_protect (uw_dispose_fd_bitmap, bitmap);
+	      add_unwind_protect (uw_dispose_command, command);	/* XXX */
 
 	      global_command = (COMMAND *)NULL;
 
 	      if ((subshell_environment & SUBSHELL_COMSUB) && comsub_ignore_return)
 		command->flags |= CMD_IGNORE_RETURN;
 
+	      if (ignore_return)
+		command->flags |= CMD_IGNORE_RETURN;
+
 #if defined (ONESHOT)
 	      /*
 	       * IF
@@ -488,13 +519,28 @@ parse_and_execute (string, from_file, flags)
 		 if we are at the end of the command string, the last in a
 		 series of connection commands is
 		 command->value.Connection->second. */
-	      else if (command->type == cm_connection && can_optimize_connection (command))
+	      else if (command->type == cm_connection &&
+		       (flags & SEVAL_NOOPTIMIZE) == 0 &&
+		       can_optimize_connection (command))
 		{
 		  command->value.Connection->second->flags |= CMD_TRY_OPTIMIZING;
 		  command->value.Connection->second->value.Simple->flags |= CMD_TRY_OPTIMIZING;
 		}
 #endif /* ONESHOT */
 
+	      /* We play tricks in the parser and command_substitute() turning
+		 expand_aliases on and off depending on which parsing pass and
+		 whether or not we're in posix mode. This only matters for
+		 parsing, and we let the higher layers deal with that. We just
+		 want to ensure that expand_aliases is set to the appropriate
+		 global value when we go to execute this command, so we save
+		 and restore it around the execution (we don't restore it if
+		 the global value of the flag (expaliases_flag) changes). */
+	      local_expalias = expand_aliases;
+	      local_alflag = expaliases_flag;
+	      if ((subshell_environment & SUBSHELL_COMSUB) || executing_funsub)
+		expand_aliases = expaliases_flag;
+
 	      /* See if this is a candidate for $( <file ). */
 	      if (startup_state == 2 &&
 		  (subshell_environment & SUBSHELL_COMSUB) &&
@@ -512,6 +558,10 @@ parse_and_execute (string, from_file, flags)
 	      dispose_fd_bitmap (bitmap);
 	      discard_unwind_frame ("pe_dispose");
 
+	      /* If the global value didn't change, we restore what we had. */
+	      if (((subshell_environment & SUBSHELL_COMSUB) || executing_funsub) && local_alflag == expaliases_flag)
+		expand_aliases = local_expalias;
+
 	      if (flags & SEVAL_ONECMD)
 		{
 		  reset_parser ();
@@ -551,6 +601,8 @@ parse_and_execute (string, from_file, flags)
       throw_to_top_level ();
     }
 
+  CHECK_TERMSIG;
+
   if (should_jump_to_top_level)
     jump_to_top_level (code);
 
@@ -563,12 +615,7 @@ parse_and_execute (string, from_file, flags)
    command substitutions during parsing to obey Posix rules about finding
    the end of the command and balancing parens. */
 int
-parse_string (string, from_file, flags, cmdp, endp)
-     char *string;
-     const char *from_file;
-     int flags;
-     COMMAND **cmdp;
-     char **endp;
+parse_string (char *string, const char *from_file, int flags, COMMAND **cmdp, char **endp)
 {
   int code, nc;
   volatile int should_jump_to_top_level;
@@ -576,6 +623,7 @@ parse_string (string, from_file, flags, cmdp, endp)
   char *ostring;
   volatile sigset_t ps_sigmask;
 
+  /* unwind-protects common to this and parse_and_execute */
   parse_prologue (string, flags, PS_TAG);
 
 #if defined (HAVE_POSIX_SIGNALS)
@@ -584,7 +632,6 @@ parse_string (string, from_file, flags, cmdp, endp)
   sigprocmask (SIG_BLOCK, (sigset_t *)NULL, (sigset_t *)&ps_sigmask);
 #endif
 
-/*itrace("parse_string: `%s'", string);*/
   /* Reset the line number if the caller wants us to.  If we don't reset the
      line number, we have to subtract one, because we will add one just
      before executing the next command (resetting the line number sets it to
@@ -692,12 +739,10 @@ out:
 }
 
 int
-open_redir_file (r, fnp)
-     REDIRECT *r;
-     char **fnp;
+open_redir_file (REDIRECT *r, char **fnp)
 {
   char *fn;
-  int fd, rval;
+  int fd;
 
   if (r->instruction != r_input_direction)
     return -1;
@@ -718,7 +763,7 @@ open_redir_file (r, fnp)
   fd = open(fn, O_RDONLY);
   if (fd < 0)
     {
-      file_error (fn);
+      internal_error ("%s: %s", fn, strerror (errno));
       free (fn);
       if (fnp)
 	*fnp = 0;
@@ -727,6 +772,9 @@ open_redir_file (r, fnp)
 
   if (fnp)
     *fnp = fn;
+  else
+    free (fn);
+
   return fd;
 }
 
@@ -734,8 +782,7 @@ open_redir_file (r, fnp)
    returning errors as appropriate, then just cats the file to the standard
    output. */
 static int
-cat_file (r)
-     REDIRECT *r;
+cat_file (REDIRECT *r)
 {
   char *fn;
   int fd, rval;
@@ -753,10 +800,7 @@ cat_file (r)
 }
 
 int
-evalstring (string, from_file, flags)
-     char *string;
-     const char *from_file;
-     int flags;
+evalstring (char *string, const char *from_file, int flags)
 {
   volatile int r, rflag, rcatch;
   volatile int was_trap;
