@@ -1,6 +1,6 @@
 /* eval.c -- reading and evaluating commands. */
 
-/* Copyright (C) 1996-2020 Free Software Foundation, Inc.
+/* Copyright (C) 1996-2024 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -48,13 +48,13 @@
 #  include "bashhist.h"
 #endif
 
-static void send_pwd_to_eterm PARAMS((void));
-static sighandler alrm_catcher PARAMS((int));
+static void send_pwd_to_eterm (void);
+static sighandler alrm_catcher (int);
 
 /* Read and execute commands until EOF is reached.  This assumes that
    the input source has already been initialized. */
 int
-reader_loop ()
+reader_loop (void)
 {
   int our_indirection_level;
   COMMAND * volatile current_command;
@@ -90,12 +90,13 @@ reader_loop ()
 	  switch (code)
 	    {
 	      /* Some kind of throw to top_level has occurred. */
-	    case FORCE_EOF:
 	    case ERREXIT:
-	    case EXITPROG:
-	      current_command = (COMMAND *)NULL;
 	      if (exit_immediately_on_error)
-		variable_context = 0;	/* not in a function */
+		reset_local_contexts ();	/* not in a function */
+	    case FORCE_EOF:
+	    case EXITPROG:
+	    case EXITBLTIN:
+	      current_command = (COMMAND *)NULL;
 	      EOF_Reached = EOF;
 	      goto exec_done;
 
@@ -139,7 +140,7 @@ reader_loop ()
 	{
 	  if (interactive_shell == 0 && read_but_dont_execute)
 	    {
-	      set_exit_status (EXECUTION_SUCCESS);
+	      set_exit_status (last_command_exit_value);
 	      dispose_command (global_command);
 	      global_command = (COMMAND *)NULL;
 	    }
@@ -152,14 +153,18 @@ reader_loop ()
 	      if (interactive && ps0_prompt)
 		{
 		  char *ps0_string;
+		  int old_eof;
 
-		  ps0_string = decode_prompt_string (ps0_prompt);
+		  old_eof = EOF_Reached;
+		  EOF_Reached = 0;
+		  ps0_string = decode_prompt_string (ps0_prompt, 1);
 		  if (ps0_string && *ps0_string)
 		    {
 		      fprintf (stderr, "%s", ps0_string);
 		      fflush (stderr);
 		    }
 		  free (ps0_string);
+		  EOF_Reached = old_eof;
 		}
 
 	      current_command_number++;
@@ -178,6 +183,11 @@ reader_loop ()
 		  current_command = (COMMAND *)NULL;
 		}
 	    }
+	  if (EOF_Reached && interactive && ignoreeof && parse_and_execute_level == 0 && code != EXITBLTIN)
+	    {
+	      if (handle_ignoreeof (1))
+		EOF_Reached = 0;
+	    }
 	}
       else
 	{
@@ -194,7 +204,7 @@ reader_loop ()
 
 /* Pretty print shell scripts */
 int
-pretty_print_loop ()
+pretty_print_loop (void)
 {
   COMMAND *current_command;
   char *command_to_print;
@@ -203,6 +213,7 @@ pretty_print_loop ()
 
   global_posix_mode = posixly_correct;
   last_was_newline = 0;
+  unset_readahead_token ();
   while (EOF_Reached == 0)
     {
       code = setjmp_nosigs (top_level);
@@ -234,11 +245,13 @@ pretty_print_loop ()
 }
 
 static sighandler
-alrm_catcher(i)
-     int i;
+alrm_catcher(int i)
 {
-  printf (_("\007timed out waiting for input: auto-logout\n"));
-  fflush (stdout);
+  char *msg;
+
+  msg = _("\007timed out waiting for input: auto-logout\n");
+  write (1, msg, strlen (msg));
+
   bash_logout ();	/* run ~/.bash_logout if this is a login shell */
   jump_to_top_level (EXITPROG);
   SIGRETURN (0);
@@ -247,7 +260,7 @@ alrm_catcher(i)
 /* Send an escape sequence to emacs term mode to tell it the
    current working directory. */
 static void
-send_pwd_to_eterm ()
+send_pwd_to_eterm (void)
 {
   char *pwd, *f;
 
@@ -260,38 +273,51 @@ send_pwd_to_eterm ()
 }
 
 #if defined (ARRAY_VARS)
+/* Caller ensures that A has a non-zero number of elements */
 int
-execute_array_command (ae, v)
-     ARRAY_ELEMENT *ae;
-     void *v;
+execute_array_command (ARRAY *a, void *v)
 {
-  char *tag, *command;
+  char *tag;
+  char **argv;
+  int argc, i;
 
   tag = (char *)v;
-  command = element_value (ae);
-  if (command && *command)
-    execute_variable_command (command, tag);
+  argc = 0;
+  argv = array_to_argv (a, &argc);
+  for (i = 0; i < argc; i++)
+    {
+      if (argv[i] && argv[i][0])
+	execute_variable_command (argv[i], tag);
+    }
+  strvec_dispose (argv);
   return 0;
 }
 #endif
   
 static void
-execute_prompt_command ()
+execute_prompt_command (void)
 {
   char *command_to_execute;
-#if defined (ARRAY_VARS)
   SHELL_VAR *pcv;
+#if defined (ARRAY_VARS)
   ARRAY *pcmds;
-
-  GET_ARRAY_FROM_VAR ("PROMPT_COMMANDS", pcv, pcmds);
-  if (pcv && var_isset (pcv) && pcmds && array_num_elements (pcmds) > 0)
-    {
-      array_walk (pcmds, execute_array_command, "PROMPT_COMMANDS");
-      return;
-    }
 #endif
 
-  command_to_execute = get_string_value ("PROMPT_COMMAND");
+  pcv = find_variable ("PROMPT_COMMAND");
+  if (pcv  == 0 || var_isset (pcv) == 0 || invisible_p (pcv))
+    return;
+#if defined (ARRAY_VARS)
+  if (array_p (pcv))
+    {
+      if ((pcmds = array_cell (pcv)) && array_num_elements (pcmds) > 0)
+	execute_array_command (pcmds, "PROMPT_COMMAND");
+      return;
+    }
+  else if (assoc_p (pcv))
+    return;	/* currently don't allow associative arrays here */
+#endif
+
+  command_to_execute = value_cell (pcv);
   if (command_to_execute && *command_to_execute)
     execute_variable_command (command_to_execute, "PROMPT_COMMAND");
 }
@@ -301,16 +327,17 @@ execute_prompt_command ()
    leaves the parsed command in the global variable GLOBAL_COMMAND.
    This is where PROMPT_COMMAND is executed. */
 int
-parse_command ()
+parse_command (void)
 {
-  int r;
+  int r, old_parsing;
 
   need_here_doc = 0;
-  run_pending_traps ();
+  if ((parser_state & (PST_CMDSUBST|PST_FUNSUBST)) == 0)
+    run_pending_traps ();
 
   /* Allow the execution of a random command just before the printing
      of each primary prompt.  If the shell variable PROMPT_COMMAND
-     is set then the value of it is the command to execute. */
+     is set then its value (array or string) is the command(s) to execute. */
   /* The tests are a combination of SHOULD_PROMPT() and prompt_again() 
      from parse.y, which are the conditions under which the prompt is
      actually printed. */
@@ -325,11 +352,14 @@ parse_command ()
 	send_pwd_to_eterm ();	/* Yuck */
     }
 
+  old_parsing = parsing_command;
+  parsing_command = 1;
   current_command_line_count = 0;
   r = yyparse ();
 
   if (need_here_doc)
     gather_here_documents ();
+  parsing_command = old_parsing;
 
   return (r);
 }
@@ -338,7 +368,7 @@ parse_command ()
    is left in the globval variable GLOBAL_COMMAND for use by reader_loop.
    This is where the shell timeout code is executed. */
 int
-read_command ()
+read_command (void)
 {
   SHELL_VAR *tmout_var;
   int tmout_len, result;
