@@ -1,6 +1,6 @@
 /* variables.h -- data structures for shell variables. */
 
-/* Copyright (C) 1987-2018 Free Software Foundation, Inc.
+/* Copyright (C) 1987-2024 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -46,6 +46,7 @@ typedef struct var_context {
 #define VC_FUNCENV	0x04	/* also function if name != NULL */
 #define VC_BLTNENV	0x08	/* builtin_env */
 #define VC_TEMPENV	0x10	/* temporary_env */
+#define VC_SPECTEMPENV	0x20	/* temporary environment preceding a posix special builtin */
 
 #define VC_TEMPFLAGS	(VC_FUNCENV|VC_BLTNENV|VC_TEMPENV)
 
@@ -61,8 +62,8 @@ typedef struct var_context {
 
 /* What a shell variable looks like. */
 
-typedef struct variable *sh_var_value_func_t __P((struct variable *));
-typedef struct variable *sh_var_assign_func_t __P((struct variable *, char *, arrayind_t, char *));
+typedef struct variable *sh_var_value_func_t (struct variable *);
+typedef struct variable *sh_var_assign_func_t (struct variable *, char *, arrayind_t, char *);
 
 /* For the future */
 union _value {
@@ -95,8 +96,8 @@ typedef struct variable {
 
 typedef struct _vlist {
   SHELL_VAR **list;
-  int list_size;	/* allocated size */
-  int list_len;		/* current number of entries */
+  size_t list_size;	/* allocated size */
+  size_t list_len;	/* current number of entries */
 } VARLIST;
 
 /* The various attributes that a given variable can have. */
@@ -114,9 +115,17 @@ typedef struct _vlist {
 #define att_capcase	0x0000400	/* word capitalized on assignment */
 #define att_nameref	0x0000800	/* word is a name reference */
 
+#define attmask_user	0x0000fff
+
 #define user_attrs	(att_exported|att_readonly|att_integer|att_local|att_trace|att_uppercase|att_lowercase|att_capcase|att_nameref)
 
-#define attmask_user	0x0000fff
+/* These define attributes you can set on readonly variables using declare.
+   You're allowed to set the readonly attribute on a readonly variable.
+   declare checks whether it gets +r explicitly, before testing these.
+   att_nameref is in there because declare performs its own validation due
+   to some ksh93 quirks. */
+#define valid_readonly_attrs	(att_exported|att_local|att_nameref|att_trace|att_readonly)
+#define invalid_readonly_attrs	(~valid_readonly_attrs & attmask_user)
 
 /* Internal attributes used for bookkeeping */
 #define att_invisible	0x0001000	/* cannot see */
@@ -162,7 +171,7 @@ typedef struct _vlist {
 /* Variable names: lvalues */
 #define name_cell(var)		((var)->name)
 
-/* Acessing variable values: rvalues */
+/* Accessing variable values: rvalues */
 #define value_cell(var)		((var)->value)
 #define function_cell(var)	(COMMAND *)((var)->value)
 #define array_cell(var)		(ARRAY *)((var)->value)
@@ -221,15 +230,17 @@ typedef struct _vlist {
 
 #define ifsname(s)	((s)[0] == 'I' && (s)[1] == 'F' && (s)[2] == 'S' && (s)[3] == '\0')
 
-/* Flag values for make_local_variable */
-#define MKLOC_INHERIT		0x01
+/* Flag values for make_local_variable and its array counterparts */
+#define MKLOC_ASSOCOK		0x01
+#define MKLOC_ARRAYOK		0x02
+#define MKLOC_INHERIT		0x04
 
 /* Special value for nameref with invalid value for creation or assignment */
 extern SHELL_VAR nameref_invalid_value;
 #define INVALID_NAMEREF_VALUE	(void *)&nameref_invalid_value
 	
 /* Stuff for hacking variables. */
-typedef int sh_var_map_func_t __P((SHELL_VAR *));
+typedef int sh_var_map_func_t (SHELL_VAR *);
 
 /* Where we keep the variables and functions */
 extern VAR_CONTEXT *global_variables;
@@ -248,200 +259,217 @@ extern int shell_level;
 
 /* XXX */
 extern WORD_LIST *rest_of_args;
+extern int posparam_count;
 extern pid_t dollar_dollar_pid;
 
-extern void initialize_shell_variables __P((char **, int));
-extern SHELL_VAR *set_if_not __P((char *, char *));
+extern int localvar_inherit;		/* declared in variables.c */
 
-extern void sh_set_lines_and_columns __P((int, int));
-extern void set_pwd __P((void));
-extern void set_ppid __P((void));
-extern void make_funcname_visible __P((int));
+extern void initialize_shell_variables (char **, int);
 
-extern SHELL_VAR *var_lookup __P((const char *, VAR_CONTEXT *));
+extern int validate_inherited_value (SHELL_VAR *, int);
 
-extern SHELL_VAR *find_function __P((const char *));
-extern FUNCTION_DEF *find_function_def __P((const char *));
-extern SHELL_VAR *find_variable __P((const char *));
-extern SHELL_VAR *find_variable_noref __P((const char *));
-extern SHELL_VAR *find_variable_last_nameref __P((const char *, int));
-extern SHELL_VAR *find_global_variable_last_nameref __P((const char *, int));
-extern SHELL_VAR *find_variable_nameref __P((SHELL_VAR *));
-extern SHELL_VAR *find_variable_nameref_for_create __P((const char *, int));
-extern SHELL_VAR *find_variable_nameref_for_assignment __P((const char *, int));
-/*extern SHELL_VAR *find_variable_internal __P((const char *, int));*/
-extern SHELL_VAR *find_variable_tempenv __P((const char *));
-extern SHELL_VAR *find_variable_notempenv __P((const char *));
-extern SHELL_VAR *find_global_variable __P((const char *));
-extern SHELL_VAR *find_global_variable_noref __P((const char *));
-extern SHELL_VAR *find_shell_variable __P((const char *));
-extern SHELL_VAR *find_tempenv_variable __P((const char *));
-extern SHELL_VAR *find_variable_no_invisible __P((const char *));
-extern SHELL_VAR *find_variable_for_assignment __P((const char *));
-extern SHELL_VAR *copy_variable __P((SHELL_VAR *));
-extern SHELL_VAR *make_local_variable __P((const char *, int));
-extern SHELL_VAR *bind_variable __P((const char *, char *, int));
-extern SHELL_VAR *bind_global_variable __P((const char *, char *, int));
-extern SHELL_VAR *bind_function __P((const char *, COMMAND *));
+extern SHELL_VAR *set_if_not (const char *, const char *);
 
-extern void bind_function_def __P((const char *, FUNCTION_DEF *, int));
+extern void sh_set_lines_and_columns (int, int);
+extern void set_pwd (void);
+extern void set_ppid (void);
+extern void make_funcname_visible (int);
 
-extern SHELL_VAR **map_over __P((sh_var_map_func_t *, VAR_CONTEXT *));
-SHELL_VAR **map_over_funcs __P((sh_var_map_func_t *));
+extern SHELL_VAR *var_lookup (const char *, VAR_CONTEXT *);
+
+extern SHELL_VAR *find_function (const char *);
+extern FUNCTION_DEF *find_function_def (const char *);
+extern SHELL_VAR *find_variable (const char *);
+extern SHELL_VAR *find_variable_noref (const char *);
+extern SHELL_VAR *find_variable_last_nameref (const char *, int);
+extern SHELL_VAR *find_global_variable_last_nameref (const char *, int);
+extern SHELL_VAR *find_variable_nameref (SHELL_VAR *);
+extern SHELL_VAR *find_variable_nameref_for_create (const char *, int);
+extern SHELL_VAR *find_variable_nameref_for_assignment (const char *, int);
+/*extern SHELL_VAR *find_variable_internal (const char *, int);*/
+extern SHELL_VAR *find_variable_tempenv (const char *);
+extern SHELL_VAR *find_variable_notempenv (const char *);
+extern SHELL_VAR *find_global_variable (const char *);
+extern SHELL_VAR *find_global_variable_noref (const char *);
+extern SHELL_VAR *find_shell_variable (const char *);
+extern SHELL_VAR *find_tempenv_variable (const char *);
+extern SHELL_VAR *find_variable_no_invisible (const char *);
+extern SHELL_VAR *find_variable_for_assignment (const char *);
+extern char *nameref_transform_name (const char *, int);
+extern SHELL_VAR *copy_variable (SHELL_VAR *);
+extern SHELL_VAR *make_local_variable (const char *, int);
+extern SHELL_VAR *bind_variable (const char *, const char *, int);
+extern SHELL_VAR *bind_global_variable (const char *, const char *, int);
+extern SHELL_VAR *bind_function (const char *, COMMAND *);
+
+extern void bind_function_def (const char *, FUNCTION_DEF *, int);
+
+extern SHELL_VAR **map_over (sh_var_map_func_t *, VAR_CONTEXT *);
+SHELL_VAR **map_over_funcs (sh_var_map_func_t *);
      
-extern SHELL_VAR **all_shell_variables __P((void));
-extern SHELL_VAR **all_shell_functions __P((void));
-extern SHELL_VAR **all_visible_variables __P((void));
-extern SHELL_VAR **all_visible_functions __P((void));
-extern SHELL_VAR **all_exported_variables __P((void));
-extern SHELL_VAR **local_exported_variables __P((void));
-extern SHELL_VAR **all_local_variables __P((void));
+extern SHELL_VAR **all_shell_variables (void);
+extern SHELL_VAR **all_shell_functions (void);
+extern SHELL_VAR **all_visible_variables (void);
+extern SHELL_VAR **all_visible_functions (void);
+extern SHELL_VAR **all_exported_variables (void);
+extern SHELL_VAR **local_exported_variables (void);
+extern SHELL_VAR **all_local_variables (int);
 #if defined (ARRAY_VARS)
-extern SHELL_VAR **all_array_variables __P((void));
+extern SHELL_VAR **all_array_variables (void);
 #endif
-extern char **all_variables_matching_prefix __P((const char *));
+extern char **all_variables_matching_prefix (const char *);
 
-extern char **make_var_array __P((HASH_TABLE *));
-extern char **add_or_supercede_exported_var __P((char *, int));
+extern char **make_var_array (HASH_TABLE *);
+extern char **add_or_supercede_exported_var (char *, int);
 
-extern char *get_variable_value __P((SHELL_VAR *));
-extern char *get_string_value __P((const char *));
-extern char *sh_get_env_value __P((const char *));
-extern char *make_variable_value __P((SHELL_VAR *, char *, int));
+extern char *get_variable_value (SHELL_VAR *);
+extern char *get_string_value (const char *);
+extern char *sh_get_env_value (const char *);
+extern char *make_variable_value (SHELL_VAR *, const char *, int);
 
-extern SHELL_VAR *bind_variable_value __P((SHELL_VAR *, char *, int));
-extern SHELL_VAR *bind_int_variable __P((char *, char *, int));
-extern SHELL_VAR *bind_var_to_int __P((char *, intmax_t));
+extern SHELL_VAR *bind_variable_value (SHELL_VAR *, char *, int);
+extern SHELL_VAR *bind_int_variable (const char *, const char *, int);
+extern SHELL_VAR *bind_var_to_int (const char *, intmax_t, int);
 
-extern int assign_in_env __P((WORD_DESC *, int));
+extern int assign_in_env (const WORD_DESC *, int);
 
-extern int unbind_variable __P((const char *));
-extern int check_unbind_variable __P((const char *));
-extern int unbind_nameref __P((const char *));
-extern int unbind_variable_noref __P((const char *));
-extern int unbind_func __P((const char *));
-extern int unbind_function_def __P((const char *));
-extern int delete_var __P((const char *, VAR_CONTEXT *));
-extern int makunbound __P((const char *, VAR_CONTEXT *));
-extern int kill_local_variable __P((const char *));
-extern void delete_all_variables __P((HASH_TABLE *));
-extern void delete_all_contexts __P((VAR_CONTEXT *));
+extern int posix_unbind_tempvar (const char *);
+extern int unbind_variable (const char *);
+extern int check_unbind_variable (const char *);
+extern int unbind_nameref (const char *);
+extern int unbind_variable_noref (const char *);
+extern int unbind_global_variable (const char *);
+extern int unbind_global_variable_noref (const char *);
+extern int unbind_func (const char *);
+extern int unbind_function_def (const char *);
+extern int delete_var (const char *, VAR_CONTEXT *);
+extern int makunbound (const char *, VAR_CONTEXT *);
+extern int kill_local_variable (const char *);
 
-extern VAR_CONTEXT *new_var_context __P((char *, int));
-extern void dispose_var_context __P((VAR_CONTEXT *));
-extern VAR_CONTEXT *push_var_context __P((char *, int, HASH_TABLE *));
-extern void pop_var_context __P((void));
-extern VAR_CONTEXT *push_scope __P((int, HASH_TABLE *));
-extern void pop_scope __P((int));
+extern void delete_all_variables (HASH_TABLE *);
+extern void delete_all_contexts (VAR_CONTEXT *);
+extern void reset_local_contexts (void);
 
-extern void push_context __P((char *, int, HASH_TABLE *));
-extern void pop_context __P((void));
-extern void push_dollar_vars __P((void));
-extern void pop_dollar_vars __P((void));
-extern void dispose_saved_dollar_vars __P((void));
+extern VAR_CONTEXT *new_var_context (char *, int);
+extern void dispose_var_context (VAR_CONTEXT *);
+extern VAR_CONTEXT *push_var_context (char *, int, HASH_TABLE *);
+extern void pop_var_context (void);
+extern VAR_CONTEXT *push_scope (int, HASH_TABLE *);
+extern void pop_scope (void *);		/* XXX uw_ */
 
-extern void init_bash_argv __P((void));
-extern void save_bash_argv __P((void));
-extern void push_args __P((WORD_LIST *));
-extern void pop_args __P((void));
+extern void clear_dollar_vars (void);
 
-extern void adjust_shell_level __P((int));
-extern void non_unsettable __P((char *));
-extern void dispose_variable __P((SHELL_VAR *));
-extern void dispose_used_env_vars __P((void));
-extern void dispose_function_env __P((void));
-extern void dispose_builtin_env __P((void));
-extern void merge_temporary_env __P((void));
-extern void flush_temporary_env __P((void));
-extern void merge_builtin_env __P((void));
-extern void kill_all_local_variables __P((void));
+extern void push_context (char *, int, HASH_TABLE *);
+extern void pop_context (void *);	/* XXX uw_ */
+extern void push_dollar_vars (void);
+extern void pop_dollar_vars (void);
+extern void dispose_saved_dollar_vars (void);
 
-extern void set_var_read_only __P((char *));
-extern void set_func_read_only __P((const char *));
-extern void set_var_auto_export __P((char *));
-extern void set_func_auto_export __P((const char *));
+extern void init_bash_argv (void);
+extern void save_bash_argv (void);
+extern void push_args (WORD_LIST *);
+extern void pop_args (void);
+extern void uw_pop_args (void *);
 
-extern void sort_variables __P((SHELL_VAR **));
+extern void adjust_shell_level (int);
+extern void non_unsettable (char *);
+extern void dispose_variable (SHELL_VAR *);
+extern void dispose_used_env_vars (void);
+extern void dispose_function_env (void);
+extern void dispose_builtin_env (void);
+extern void merge_temporary_env (void);
+extern void merge_function_temporary_env (void);
+extern void flush_temporary_env (void);
+extern void merge_builtin_env (void);
+extern void kill_all_local_variables (void);
 
-extern int chkexport __P((char *));
-extern void maybe_make_export_env __P((void));
-extern void update_export_env_inplace __P((char *, int, char *));
-extern void put_command_name_into_env __P((char *));
-extern void put_gnu_argv_flags_into_env __P((intmax_t, char *));
+extern void set_var_read_only (char *);
+extern void set_func_read_only (const char *);
+extern void set_var_auto_export (char *);
+extern void set_func_auto_export (const char *);
 
-extern void print_var_list __P((SHELL_VAR **));
-extern void print_func_list __P((SHELL_VAR **));
-extern void print_assignment __P((SHELL_VAR *));
-extern void print_var_value __P((SHELL_VAR *, int));
-extern void print_var_function __P((SHELL_VAR *));
+extern void sort_variables (SHELL_VAR **);
+
+extern int chkexport (char *);
+extern void maybe_make_export_env (void);
+extern void update_export_env_inplace (char *, int, char *);
+extern void put_command_name_into_env (char *);
+extern void put_gnu_argv_flags_into_env (intmax_t, char *);
+
+extern void print_var_list (SHELL_VAR **);
+extern void print_func_list (SHELL_VAR **);
+extern void print_assignment (SHELL_VAR *);
+extern void print_var_value (SHELL_VAR *, int);
+extern void print_var_function (SHELL_VAR *);
 
 #if defined (ARRAY_VARS)
-extern SHELL_VAR *make_new_array_variable __P((char *));
-extern SHELL_VAR *make_local_array_variable __P((char *, int));
+extern SHELL_VAR *make_new_array_variable (const char *);
+extern SHELL_VAR *make_local_array_variable (const char *, int);
 
-extern SHELL_VAR *make_new_assoc_variable __P((char *));
-extern SHELL_VAR *make_local_assoc_variable __P((char *));
+extern SHELL_VAR *make_new_assoc_variable (const char *);
+extern SHELL_VAR *make_local_assoc_variable (const char *, int);
 
-extern void set_pipestatus_array __P((int *, int));
-extern ARRAY *save_pipestatus_array __P((void));
-extern void restore_pipestatus_array __P((ARRAY *));
+extern void set_pipestatus_array (int *, int);
+extern ARRAY *save_pipestatus_array (void);
+extern void restore_pipestatus_array (ARRAY *);
 #endif
 
-extern void set_pipestatus_from_exit __P((int));
+extern void set_pipestatus_from_exit (int);
 
 /* The variable in NAME has just had its state changed.  Check to see if it
    is one of the special ones where something special happens. */
-extern void stupidly_hack_special_variables __P((char *));
+extern void stupidly_hack_special_variables (const char *);
 
 /* Reinitialize some special variables that have external effects upon unset
    when the shell reinitializes itself. */
-extern void reinit_special_variables __P((void));
+extern void reinit_special_variables (void);
 
-extern int get_random_number __P((void));
+extern int get_random_number (void);
 
 /* The `special variable' functions that get called when a particular
    variable is set. */
-extern void sv_ifs __P((char *));
-extern void sv_path __P((char *));
-extern void sv_mail __P((char *));
-extern void sv_funcnest __P((char *));
-extern void sv_execignore __P((char *));
-extern void sv_globignore __P((char *));
-extern void sv_ignoreeof __P((char *));
-extern void sv_strict_posix __P((char *));
-extern void sv_optind __P((char *));
-extern void sv_opterr __P((char *));
-extern void sv_locale __P((char *));
-extern void sv_xtracefd __P((char *));
-extern void sv_shcompat __P((char *));
+extern void sv_ifs (const char *);
+extern void sv_path (const char *);
+extern void sv_mail (const char *);
+extern void sv_funcnest (const char *);
+extern void sv_execignore (const char *);
+extern void sv_globignore (const char *);
+extern void sv_ignoreeof (const char *);
+extern void sv_strict_posix (const char *);
+extern void sv_optind (const char *);
+extern void sv_opterr (const char *);
+extern void sv_locale (const char *);
+extern void sv_xtracefd (const char *);
+extern void sv_shcompat (const char *);
+extern void sv_globsort (const char *);
 
 #if defined (READLINE)
-extern void sv_comp_wordbreaks __P((char *));
-extern void sv_terminal __P((char *));
-extern void sv_hostfile __P((char *));
-extern void sv_winsize __P((char *));
+extern void sv_comp_wordbreaks (const char *);
+extern void sv_terminal (const char *);
+extern void sv_hostfile (const char *);
+extern void sv_winsize (const char *);
 #endif
 
 #if defined (__CYGWIN__)
-extern void sv_home __P((char *));
+extern void sv_home (const char *);
 #endif
 
 #if defined (HISTORY)
-extern void sv_histsize __P((char *));
-extern void sv_histignore __P((char *));
-extern void sv_history_control __P((char *));
+extern void sv_histsize (const char *);
+extern void sv_histignore (const char *);
+extern void sv_history_control (const char *);
 #  if defined (BANG_HISTORY)
-extern void sv_histchars __P((char *));
+extern void sv_histchars (const char *);
 #  endif
-extern void sv_histtimefmt __P((char *));
+extern void sv_histtimefmt (const char *);
 #endif /* HISTORY */
 
 #if defined (HAVE_TZSET)
-extern void sv_tz __P((char *));
+extern void sv_tz (const char *);
 #endif
 
 #if defined (JOB_CONTROL)
-extern void sv_childmax __P((char *));
+extern void sv_childmax (const char *);
 #endif
 
 #endif /* !_VARIABLES_H_ */
